"""Defines an iterator to loop through VT API collections."""

import typing

from .object import Object
from .utils import make_sync

if typing.TYPE_CHECKING:
    from .client import Client

__all__ = ["Iterator"]


class Iterator:
    # pylint: disable=line-too-long
    """Iterator allows iterating over object collections.

    Some endpoints in the VirusTotal API represent a collection of objects, for
    example:

    `/files/{id}/comments <https://docs.virustotal.com/reference/get-comments>`_

    `/intelligence/search <https://docs.virustotal.com/reference/search>`_

    These collections can be iterated using an instance of this class.

    Learn more about collections in the VirusTotal API in:
    https://docs.virustotal.com/reference/collections-create

    The following example iterates over the most recent 200 comments, retrieving
    them in batches of 20:

    >>> client = vt.Client(<apikey>)
    >>> it = client.iterator('/comments', batch_size=20, limit=200)
    >>> for comment in it:
    >>>   print(comment.text)
    >>> print(it.cursor)

    When the iteration is done, it print the iterator's cursor. The cursor can be
    used for creating another iterator that continues at the point where the
    previous iterator left.

    The Iterator class also exposes an async iterator:

    >>>  # Define an async coroutine that iterates over the comments.
    >>>  async def print_comments():
    >>>    async for comment in client.iterator('/comments', limit=200):
    >>>      print(comment.id)
    >>>  # Run the print_comments coroutine using asyncio
    >>>  import asyncio
    >>>  asyncio.get_event_loop().run_until_complete(print_comments)
    """

    # pylint: disable=line-too-long

    def __init__(
            self,
            client: "Client",
            path: str, params=None,
            cursor: typing.Optional[str] = None,
            limit: typing.Optional[int] = None,
            batch_size: int = 0
    ):
        """Initializes an iterator.

        This function is not intended to be called directly. Client.iterator() is
        the preferred way for creating an iteraror.
        """
        self._client = client
        self._path = path
        self._params: typing.Dict = params or {}
        self._batch_size = batch_size
        self._limit = limit
        self._items: typing.List[typing.Dict] = []
        self._count = 0
        self._server_cursor: typing.Optional[str] = None
        self._batch_cursor = 0
        self._meta: typing.Optional[typing.Dict] = None

        if "cursor" in self._params:
            raise ValueError('Do not pass "cursor" as a path param')

        if "limit" in self._params:
            raise ValueError('Do not pass "limit" as a path param')

        if cursor:
            self._server_cursor, _, batch_cursor = cursor.rpartition("-")
            if not self._server_cursor:
                raise ValueError("invalid cursor")
            try:
                self._batch_cursor = int(batch_cursor)
            except ValueError as exc:
                raise ValueError("invalid cursor") from exc

    def _build_params(self) -> typing.Dict:
        params = self._params.copy()
        if self._server_cursor:
            params["cursor"] = self._server_cursor
        if self._batch_size:
            params["limit"] = self._batch_size
        return params

    def _parse_response(self, json_resp: typing.Dict, batch_cursor: int) -> typing.Tuple[
        typing.List[typing.Dict], typing.Dict]:
        if not isinstance(json_resp.get("data"), list):
            raise ValueError(f"{self._path} is not a collection")
        meta = json_resp.get("meta", {})
        items = json_resp["data"][batch_cursor:]

        return items, meta

    async def _get_batch_async(self, batch_cursor: int = 0) -> typing.Tuple[typing.List[typing.Dict], typing.Dict]:
        json_resp = await self._client.get_json_async(
            self._path, params=self._build_params()
        )
        return self._parse_response(json_resp, batch_cursor)

    def __iter__(self):
        return self

    def __aiter__(self):
        return self

    def __next__(self) -> Object:
        try:
            return make_sync(self.__anext__())
        except StopAsyncIteration as exc:
            raise StopIteration() from exc

    async def __anext__(self) -> Object:
        if self._limit and self._count == self._limit:
            raise StopAsyncIteration()
        if not self._items and (self.cursor or self._count == 0):
            self._items, self._meta = await self._get_batch_async()
            self._server_cursor = self._meta.pop("cursor", None)
            self._batch_cursor = 0
        if not self._items and not self._server_cursor:
            raise StopAsyncIteration()
        item = self._items.pop(0)
        self._count += 1
        self._batch_cursor += 1
        return Object.from_dict(item)

    @property
    def cursor(self) -> typing.Optional[str]:
        """Cursor indicating the last returned object.

        This cursor can be used for creating a new iterator that continues where
        the current one left.
        """
        if not self._server_cursor or not self._count:
            return None
        return self._server_cursor + "-" + str(self._batch_cursor)

    @property
    async def meta_async(self) -> typing.Dict:
        """Meta information.

        The cursor is not included, as it's exposed as a property.
        """
        if self._meta is None:
            # Load the first batch of items in order to retrieve the meta info.
            self._items, self._meta = await self._get_batch_async()
            self._server_cursor = self._meta.pop("cursor", None)
            self._batch_cursor = 0
        return self._meta

    @property
    def meta(self) -> typing.Dict:
        return make_sync(self.meta_async)
